"""
This module 'populate' defines methods to populate the database with predefined ('static') data,
according to the proposed Django way: https://docs.djangoproject.com/en/2.0/topics/migrations/#data-migrations

import this module in your empty migration step file, and add the following migration:

from ..populate import *

class Migration(migrations.Migration):

    dependencies = [ <the dependency is automatically inserted here> ]

    operations = [ migrations.RunPython(populate_choices) ]

"""

import json
from lofar.sas.tmss.tmss.tmssapp.models.specification import Role, Datatype, Dataformat, CopyReason, TaskTemplate, ResourceType, ResourceUnit
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import SubtaskState, SubtaskType, SubtaskTemplate, Subtask, \
    StationType, Algorithm, ScheduleMethod, Cluster, Filesystem
from lofar.common.json_utils import *

def populate_choices(apps, schema_editor):
    '''
    populate each 'choice' table in the database with the 'static' list of 'choice'.Choices for
    each 'choice'type in Role, Datatype, Dataformat, CopyReason
    :return: None
    '''
    for choice_class in [Role, Datatype, Dataformat, CopyReason,
                         SubtaskState, SubtaskType, StationType, Algorithm, ScheduleMethod]:
        choice_class.objects.bulk_create([choice_class(value=x.value) for x in choice_class.Choices])


def populate_lofar_json_schemas(apps, schema_editor):
    _populate_correlator_calibrator_schema()
    _populate_obscontrol_schema()
    _populate_stations_schema()

    _populate_example_data()

def populate_resources(apps, schema_editor):
    ru_bytes = ResourceUnit.objects.create(name="bytes", description="Bytes")
    ru_hours = ResourceUnit.objects.create(name="hours", description="duration in hours")

    ResourceType.objects.create(name="lta_storage", description="Amount of storage in LTA", resource_unit=ru_bytes)
    ResourceType.objects.create(name="cep_storage", description="Amount of storage at CEP processing cluster", resource_unit=ru_bytes)
    ResourceType.objects.create(name="cep_processing_hours", description="Number of processing hours for CEP processing cluster", resource_unit=ru_hours)

def populate_misc(apps, schema_editor):
    cluster = Cluster.objects.create(name="CEP4", location="CIT")
    fs = Filesystem.objects.create(name="LustreFS", cluster=cluster, capacity=3.6e15)


def _populate_correlator_calibrator_schema():
    task_template_data = {"name": "correlator schema",
                          "description": 'correlator schema for calibrator observations',
                          "version": '0.1',
                          "schema": json.loads('''
{
  "$id": "http://example.com/example.json",
  "type": "object",
  "$schema": "http://json-schema.org/draft-06/schema#",
  "definitions": {
    "pointing": {
      "type": "object",
      "additionalProperties": false,
      "properties": {
        "direction_type": {
          "type": "string",
          "title": "Reference frame",
          "description": "",
          "default": "J2000",
          "enum": [
            "J2000",
            "SUN",
            "MOON",
            "MERCURY",
            "VENUS",
            "MARS",
            "JUPITER",
            "SATURN",
            "URANUS",
            "NEPTUNE",
            "PLUTO"
          ]
        },
        "angle1": {
          "type": "number",
          "title": "Angle 1",
          "description": "First angle (f.e. RA)",
          "default": 0
        },
        "angle2": {
          "type": "number",
          "title": "Angle 2",
          "description": "Second angle (f.e. DEC)",
          "default": 0
        }
      }
    }
  },
  "additionalProperties": false,
  "properties": {
    "duration": {
      "type": "number",
      "title": "Duration (seconds)",
      "description": "Duration of this observation",
      "default": 60,
      "minimum": 1
    },
    "calibrator": {
      "type": "object",
      "additionalProperties": false,
      "default": {},
      "properties": {
        "enabled": {
          "type": "boolean",
          "title": "Calibrator",
          "description": "Replace targets by calibrators",
          "default": false
        },
        "autoselect": {
          "type": "boolean",
          "title": "Auto-select",
          "description": "Auto-select calibrator based on elevation",
          "default": false
        },
        "pointing": {
          "title": "Digital pointing",
          "$ref": "#/definitions/pointing",
          "default": {}
        }
      }
    },
    "channels_per_subband": {
      "type": "integer",
      "title": "Channels/subband",
      "description": "Number of frequency bands per subband",
      "default": 64,
      "minimum": 8,
      "enum": [
        8,
        16,
        32,
        64,
        128,
        256,
        512,
        1024
      ]
    },
    "integration_time": {
      "type": "number",
      "title": "Integration time (seconds)",
      "description": "Desired integration period",
      "default": 1,
      "minimum": 0.1
    },
    "storage_cluster": {
      "type": "string",
      "title": "Storage cluster",
      "description": "Cluster to write output to",
      "default": "CEP4",
      "enum": [
        "CEP4",
        "DragNet"
      ]
    }
  }
}'''), "tags": []}

    TaskTemplate.objects.create(**task_template_data)


def _populate_example_data():
    try:
        from datetime import datetime
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.test.tmss_test_data_django_models import TaskDraft_test_data, TaskBlueprint_test_data, SubtaskOutput_test_data, Dataproduct_test_data, Subtask_test_data

        task_template = models.TaskTemplate.objects.get(name='correlator schema')
        task_draft_data = TaskDraft_test_data(name="my test obs", specifications_template=task_template)
        task_draft = models.TaskDraft.objects.create(**task_draft_data)

        task_blueprint_data = TaskBlueprint_test_data(task_draft=task_draft)
        task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)

        subtask_template = models.SubtaskTemplate.objects.get(name='observationcontrol schema')
        specifications_doc = {
            "stations": {"station_list": ["CS021"],
                         "antenna_set": "HBA_DUAL_INNER",
                         "filter": "HBA_110_190",
                         "analog_pointing": {"direction_type": "J2000",
                                             "angle1": 0.4262457643630986,
                                             "angle2": 0.5787463318245085 },
                         "digital_pointings": [{"name": "beam01",
                                                "pointing": {"direction_type": "J2000",
                                                             "angle1": 0.4262457643630986,
                                                             "angle2": 0.5787463318245085 },
                                                "subbands": list(range(0, 244))
                                                }]
                         }
        }

        specifications_doc = add_defaults_to_json_object_for_schema(specifications_doc, subtask_template.schema)
        subtask_data = Subtask_test_data(task_blueprint=task_blueprint, subtask_template=subtask_template, specifications_doc=specifications_doc)
        subtask = models.Subtask.objects.create(**subtask_data)

        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))
        for sb_nr in specifications_doc['stations']['digital_pointings'][0]['subbands']:
            models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output,
                                                                      directory="CEP4:/data/test-projects/L%d/uv/" % (subtask.id,),
                                                                      filename="L%d_SB%03d_uv.MS"%(subtask.id, sb_nr)))
    except ImportError:
        pass

def _populate_obscontrol_schema():
    subtask_template_data = {"type": SubtaskType.objects.get(value='observation'),
                             "name": "observationcontrol schema",
                             "description": 'observationcontrol schema for observation subtask',
                             "version": '0.1',
                             "schema": json.loads('''
{
  "$id": "http://example.com/example.json",
  "type": "object",
  "$schema": "http://json-schema.org/draft-06/schema#",
  "definitions": {
    "pointing": {
      "type": "object",
      "additionalProperties": false,
      "properties": {
        "direction_type": {
          "type": "string",
          "title": "Reference frame",
          "description": "",
          "default": "J2000",
          "enum": [
            "J2000",
            "SUN",
            "MOON",
            "MERCURY",
            "VENUS",
            "MARS",
            "JUPITER",
            "SATURN",
            "URANUS",
            "NEPTUNE",
            "PLUTO"
          ]
        },
        "angle1": {
          "type": "number",
          "title": "Angle 1",
          "description": "First angle (f.e. RA)",
          "default": 0
        },
        "angle2": {
          "type": "number",
          "title": "Angle 2",
          "description": "Second angle (f.e. DEC)",
          "default": 0
        }
      }
    }
  },
  "additionalProperties": false,
  "properties": {
    "stations": {
      "type": "object",
      "default": {},
      "properties": {
          "station_list": {
            "title": "Station list",
            "type": "array",
            "additionalItems": false,
            "additionalProperties": false,
            "default": ["CS001"],
            "items": {
              "type": "string",
              "enum": [
                "CS001",
                "CS002",
                "CS003",
                "CS004",
                "CS005",
                "CS006",
                "CS007",
                "CS011",
                "CS013",
                "CS017",
                "CS021",
                "CS024",
                "CS026",
                "CS028",
                "CS030",
                "CS031",
                "CS032",
                "CS101",
                "CS103",
                "CS201",
                "CS301",
                "CS302",
                "CS401",
                "CS501",
                "RS106",
                "RS205",
                "RS208",
                "RS210",
                "RS305",
                "RS306",
                "RS307",
                "RS310",
                "RS406",
                "RS407",
                "RS409",
                "RS503",
                "RS508",
                "RS509",
                "DE601",
                "DE602",
                "DE603",
                "DE604",
                "DE605",
                "FR606",
                "SE607",
                "UK608",
                "DE609",
                "PL610",
                "PL611",
                "PL612",
                "IE613",
                "LV614"
              ],
              "title": "Station",
              "description": ""
            },
            "minItems": 1,
            "uniqueItems": true
          },
          "antenna_set": {
            "type": "string",
            "title": "Antenna set",
            "description": "Which antennas & fields to use on each station",
            "default": "HBA_DUAL",
            "enum": [
              "HBA_DUAL",
              "HBA_DUAL_INNER",
              "HBA_ONE",
              "HBA_ONE_INNER",
              "HBA_ZERO",
              "HBA_ZERO_INNER",
              "LBA_INNER",
              "LBA_OUTER",
              "LBA_SPARSE_EVEN",
              "LBA_SPARSE_ODD",
              "LBA_ALL"
            ]
          },
          "filter": {
            "type": "string",
            "title": "Band-pass filter",
            "description": "Must match antenna type",
            "default": "HBA_110_190",
            "enum": [
              "LBA_10_90",
              "LBA_30_90",
              "HBA_110_190",
              "HBA_210_250"
            ]
          },
          "analog_pointing": {
            "title": "Analog pointing",
            "description": "HBA only",
            "$ref": "#/definitions/pointing",
            "default": {}
          },
          "digital_pointings": {
            "type": "array",
            "title": "Beams",
            "additionalItems": false,
            "default": [{}],
            "items": {
              "title": "Beam",
              "headerTemplate": "{{ i0 }} - {{ self.name }}",
              "type": "object",
              "additionalProperties": false,
              "properties": {
                "name": {
                  "type": "string",
                  "title": "Name/target",
                  "description": "Custom identifier for this beam. Same name is same beam.",
                  "default": ""
                },
                "pointing": {
                  "title": "Digital pointing",
                  "$ref": "#/definitions/pointing",
                  "default": {}
                },
                "subbands": {
                  "type": "array",
                  "title": "Subband list",
                  "additionalItems": false,
                  "default": [],
                  "items": {
                    "type": "integer",
                    "title": "Subband",
                    "minimum": 0,
                    "maximum": 511
                  }
                }
              }
            }
        }
      }
    },
    "COBALT": {
      "type": "object",
      "title": "COBALT correlator/beamformer",
      "additionalProperties": false,
      "default": {},
      "properties": {
        "blocksize": {
          "type": "integer",
          "title": "Block size (samples)",
          "description": "Size of blocks COBALT works on, must be a multiple of all processing requirements",
          "default": 196608,
          "minimum": 97656,
          "maximum": 292968
        },
        "delay_compensation": {
          "type": "boolean",
          "title": "Apply delay compensation",
          "description": "Compensate for geometric and clock differences",
          "default": true
        },
        "bandpass_correction": {
          "type": "boolean",
          "title": "Apply band-pass correction",
          "description": "Compensate for differences in station sensitivity within a subband",
          "default": true
        },
        "correlator": {
          "title": "Correlator",
          "type": "object",
          "default": {},
          "oneOf": [
            {
              "type": "object",
              "title": "Enabled",
              "additionalProperties": false,
              "default": {},
              "properties": {
                "enabled": {
                  "type": "boolean",
                  "title": "Enabled",
                  "description": "",
                  "default": true,
                  "options": {
                    "hidden": true
                  },
                  "enum": [
                    true
                  ]
                },
                "channels_per_subband": {
                  "type": "integer",
                  "title": "Channels/subband",
                  "description": "Number of frequency bands per subband",
                  "default": 64,
                  "minimum": 1,
                  "enum": [
                    1,
                    8,
                    16,
                    32,
                    64,
                    128,
                    256,
                    512,
                    1024
                  ]
                },
                "blocks_per_integration": {
                  "type": "integer",
                  "title": "Blocks per integration",
                  "description": "Number of blocks to integrate",
                  "default": 1,
                  "minimum": 1
                },
                "integrations_per_block": {
                  "type": "integer",
                  "title": "Integrations per block",
                  "description": "Number of integrations to fit within each block",
                  "default": 1,
                  "minimum": 1
                },
                "phase_centers": {
                  "type": "array",
                  "title": "Custom phase centers",
                  "additionalItems": false,
                  "default": [{}],
                  "items": {
                    "title": "Beam",
                    "headerTemplate": "Beam {{ self.index }}",
                    "type": "object",
                    "additionalProperties": false,
                    "default": {},
                    "properties": {
                      "index": {
                        "type": "integer",
                        "title": "Station beam index",
                        "description": "Apply to this station beam",
                        "minimum": 0,
                        "default": 0
                      },
                      "pointing": {
                        "title": "Correlator pointing",
                        "$ref": "#/definitions/pointing",
                        "default": {}
                      }
                    }
                  }
                }
              }
            },
            {
              "type": "object",
              "title": "Disabled",
              "additionalProperties": false,
              "default": {},
              "properties": {}
            }
          ]
        }
      }
    }
  }
}'''),
                             "realtime": True,
                             "queue": False,
                             "tags": []}

    SubtaskTemplate.objects.create(**subtask_template_data)


def _populate_stations_schema():
    task_template_data = { "name": "stations schema",
                           "description": 'Generic station settings and selection',
                           "version": '0.1',
                           "schema": json.loads('''
{
    "$id": "http://example.com/example.json",
    "type": "object",
    "$schema": "http://json-schema.org/draft-06/schema#",
    "definitions": {
        "pointing": {
            "type": "object",
            "additionalProperties": false,
            "properties": {
                "direction_type": {
                    "type": "string",
                    "title": "Reference frame",
                    "description": "",
                    "default": "J2000",
                    "enum": [
                        "J2000",
                        "SUN",
                        "MOON",
                        "MERCURY",
                        "VENUS",
                        "MARS",
                        "JUPITER",
                        "SATURN",
                        "URANUS",
                        "NEPTUNE",
                        "PLUTO"
                    ]
                },
                "angle1": {
                    "type": "number",
                    "title": "Angle 1",
                    "description": "First angle (f.e. RA)",
                    "default": 0
                },
                "angle2": {
                    "type": "number",
                    "title": "Angle 2",
                    "description": "Second angle (f.e. DEC)",
                    "default": 0
                }
            }
        }
    },
    "additionalProperties": false,
    "properties": {
        "stations": {
            "title": "Station list",
            "type": "array",
            "default": [{}],
            "oneOf": [{
                    "type": "array",
                    "title": "Fixed list",
                    "additionalItems": false,
                    "additionalProperties": false,
                    "default": ["CS001"],
                    "items": {
                        "type": "string",
                        "enum": [
                            "CS001",
                            "CS002",
                            "CS003",
                            "CS004",
                            "CS005",
                            "CS006",
                            "CS007",
                            "CS011",
                            "CS013",
                            "CS017",
                            "CS021",
                            "CS024",
                            "CS026",
                            "CS028",
                            "CS030",
                            "CS031",
                            "CS032",
                            "CS101",
                            "CS103",
                            "CS201",
                            "CS301",
                            "CS302",
                            "CS401",
                            "CS501",
                            "RS104",
                            "RS106",
                            "RS205",
                            "RS208",
                            "RS210",
                            "RS305",
                            "RS306",
                            "RS307",
                            "RS310",
                            "RS406",
                            "RS407",
                            "RS409",
                            "RS410",
                            "RS503",
                            "RS508",
                            "RS509",
                            "DE601",
                            "DE602",
                            "DE603",
                            "DE604",
                            "DE605",
                            "FR606",
                            "SE607",
                            "UK608",
                            "DE609",
                            "PL610",
                            "PL611",
                            "PL612",
                            "IE613",
                            "LV614"
                        ],
                        "title": "Station",
                        "description": ""
                    },
                    "minItems": 1,
                    "uniqueItems": true
                },
                {
                    "title": "Dynamic list",
                    "type": "array",
                    "additionalItems": false,
                    "default": [{}],
                    "items": {
                        "type": "object",
                        "title": "Station set",
                        "headerTemplate": "{{ self.group }}",
                        "additionalProperties": false,
                        "default": {},
                        "properties": {
                            "group": {
                                "type": "string",
                                "title": "Group/station",
                                "description": "Which (group of) station(s) to select from",
                                "default": "ALL",
                                "enum": [
                                    "ALL",
                                    "SUPERTERP",
                                    "CORE",
                                    "REMOTE",
                                    "DUTCH",
                                    "INTERNATIONAL"
                                ]
                            },
                            "min_stations": {
                                "type": "integer",
                                "title": "Minimum nr of stations",
                                "description": "Number of stations to use within group/station",
                                "default": 1,
                                "minimum": 0
                            }
                        }
                    }
                }
            ]
        },
        "antenna_set": {
            "type": "string",
            "title": "Antenna set",
            "description": "Fields & antennas to use",
            "default": "HBA_DUAL",
            "enum": [
                "HBA_DUAL",
                "HBA_DUAL_INNER",
                "HBA_JOINED",
                "HBA_JOINED_INNER",
                "HBA_ONE",
                "HBA_ONE_INNER",
                "HBA_ZERO",
                "HBA_ZERO_INNER",
                "LBA_INNER",
                "LBA_OUTER",
                "LBA_SPARSE_EVEN",
                "LBA_SPARSE_ODD",
                "LBA_ALL"
            ]
        },
        "filter": {
            "type": "string",
            "title": "Band-pass filter",
            "description": "Must match antenna type",
            "default": "HBA_110_190",
            "enum": [
                "LBA_10_70",
                "LBA_30_70",
                "LBA_10_90",
                "LBA_30_90",
                "HBA_110_190",
                "HBA_210_250"
            ]
        },
        "analog_pointing": {
            "title": "Analog pointing",
            "description": "HBA only",
            "default": {},
            "$ref": "#/definitions/pointing"
        },
        "beams": {
            "type": "array",
            "title": "Beams",
            "additionalItems": false,
            "default": [{}],
            "items": {
                "title": "Beam",
                "headerTemplate": "{{ i0 }} - {{ self.name }}",
                "type": "object",
                "additionalProperties": false,
                "default": {},
                "properties": {
                    "name": {
                        "type": "string",
                        "title": "Name/target",
                        "description": "Identifier for this beam",
                        "default": ""
                    },
                    "digital_pointing": {
                        "title": "Digital pointing",
                        "default": {},
                        "$ref": "#/definitions/pointing"
                    },
                    "subbands": {
                        "type": "array",
                        "title": "Subband list",
                        "additionalItems": false,
                        "default": [],
                        "items": {
                            "type": "integer",
                            "title": "Subband",
                            "minimum": 0,
                            "maximum": 511
                        }
                    }
                }
            }
        }
    }
}'''),
                           "tags": []}

    TaskTemplate.objects.create(**task_template_data)
