"""
This file contains the serializers (for the elsewhere defined data models)
"""

from rest_framework import serializers
from .. import models
from django.contrib.auth.models import User
from rest_framework import decorators

class RelationalHyperlinkedModelSerializer(serializers.HyperlinkedModelSerializer):

    def get_field_names(self, declared_fields, info):
        expanded_fields = super(RelationalHyperlinkedModelSerializer, self).get_field_names(declared_fields, info)

        if getattr(self.Meta, 'extra_fields', None):
            return expanded_fields + self.Meta.extra_fields
        else:
            return expanded_fields


# This is required for keeping a user reference as ForeignKey in other models
# (I think so that the HyperlinkedModelSerializer can generate a URI)
class UserSerializer(serializers.Serializer):
    class Meta:
        model = User
        fields = '__all__'


class TagsSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Tags
        fields = '__all__'


class GeneratorTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.GeneratorTemplate
        fields = '__all__'


class DefaultGeneratorTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultGeneratorTemplate
        fields = '__all__'


class SchedulingUnitTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.SchedulingUnitTemplate
        fields = '__all__'


class DefaultSchedulingUnitTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultSchedulingUnitTemplate
        fields = '__all__'


class TaskTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.TaskTemplate
        fields = '__all__'


class DefaultTaskTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultTaskTemplate
        fields = '__all__'


class WorkRelationSelectionTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.WorkRelationSelectionTemplate
        fields = '__all__'


class DefaultWorkRelationSelectionTemplateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultWorkRelationSelectionTemplate
        fields = '__all__'


class RoleSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Role
        fields = '__all__'


class DatatypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Datatype
        fields = '__all__'


class DataformatSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Dataformat
        fields = '__all__'


class CopyReasonSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.CopyReason
        fields = '__all__'

        

class TaskConnectorsSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.TaskConnectors
        fields = '__all__'


class CycleSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Cycle
        fields = '__all__'
        extra_fields = ['projects', 'name']


class ProjectSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Project
        fields = '__all__'
        extra_fields = ['name','project_quota']

class ProjectQuotaSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ProjectQuota
        fields = '__all__'
        extra_fields = ['resource_type']

class ResourceUnitSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ResourceUnit
        fields = '__all__'
        extra_fields = ['name']

class ResourceTypeSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ResourceType
        fields = '__all__'
        extra_fields = ['name']

class SchedulingSetSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SchedulingSet
        fields = '__all__'
        extra_fields = ['scheduling_unit_drafts']


class SchedulingUnitDraftSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SchedulingUnitDraft
        fields = '__all__'
        extra_fields = ['related_scheduling_unit_blueprint', 'task_drafts']


class SchedulingUnitBlueprintSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.SchedulingUnitBlueprint
        fields = '__all__'


class TaskDraftSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskDraft
        fields = '__all__'
        extra_fields = ['related_task_blueprint', 'produced_by', 'consumed_by']


class TaskBlueprintSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskBlueprint
        fields = '__all__'
        extra_fields = ['subtasks', 'produced_by', 'consumed_by']


class TaskRelationDraftSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskRelationDraft
        fields = '__all__'
        extra_fields = ['related_task_relation_blueprint']


class TaskRelationBlueprintSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.TaskRelationBlueprint
        fields = '__all__'


# ----- JSON

class TaskBlueprintSerializerJSONeditorOnline(RelationalHyperlinkedModelSerializer):

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # blueprint refers to. If that fails, the JSONField remains a standard text input.
    #
    # Note: I feel a bit uneasy with this since I feel there should be a more straight-forward solution tham
    #  ...intercepting the init process to determine the schema (or template uri or so) for the style attribute.
    #  ...Hoewever, I did not manage to simply pass the value(!) of e.g. the template field as a style attribute
    #  ...of the JSONField via a SerializerMethodField or similar, although I feel that should be possible.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        import json

        try:
            schema = self.instance.specifications_template.schema
            self.fields['specifications_doc'] = serializers.JSONField(
                style={'template': 'josdejong_jsoneditor_widget.html',
                       'schema': json.dumps(schema)})

        except:
            # todo: Shall we use one of the default templates for the init?
            print('Could not determine schema, hence no fancy JSON form. Expected for list view.')

    class Meta:
        model = models.TaskBlueprint
        fields = '__all__'
        extra_fields = ['subtasks', 'produced_by', 'consumed_by']


class TaskDraftSerializerJSONeditorOnline(RelationalHyperlinkedModelSerializer):

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    #
    # Note: I feel a bit uneasy with this since I feel there should be a more straight-forward solution tham
    #  ...intercepting the init process to determine the schema (or template uri or so) for the style attribute.
    #  ...Hoewever, I did not manage to simply pass the value(!) of e.g. the template field as a style attribute
    #  ...of the JSONField via a SerializerMethodField or similar, although I feel that should be possible.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        import json

        try:
            schema = self.instance.specifications_template.schema
            self.fields['specifications_doc'] = serializers.JSONField(
                style={'template': 'josdejong_jsoneditor_widget.html',
                       'schema': json.dumps(schema)})

        except:
            # todo: Shall we use one of the default templates for the init?
            print('Could not determine schema, hence no fancy JSON form. Expected for list view.')

    class Meta:
        model = models.TaskDraft
        fields = '__all__'
        extra_fields = ['related_task_blueprint', 'produced_by', 'consumed_by']


